from .helpers import MockerTestCase
from passpie.utils import genpass, mkdir_open


class UtilsTests(MockerTestCase):

    def test_genpass_generates_a_password_with_length_32(self):
        password = genpass()
        self.assertEqual(len(password), 32)

    def test_mkdir_open_makedirs_on_path_dirname(self):
        mock_os = self.patch("passpie.utils.os")
        self.patch("passpie.utils.open", self.mock_open(), create=True)
        path = "path/to/foo.pass"
        with mkdir_open(path, "w"):
            dirname = mock_os.path.dirname(path)
            mock_os.makedirs.assert_called_once_with(dirname)

    def test_mkdir_open_handle_oserror_for_file_exist(self):
        mock_os = self.patch("passpie.utils.os")
        self.patch("passpie.utils.open", self.mock_open(), create=True)
        path = "path/to/foo.pass"

        mock_os.makedirs.side_effect = OSError(17, "File Exists")
        with mkdir_open(path, "w") as fd:
            self.assertIsNotNone(fd)

        mock_os.makedirs.side_effect = OSError(2, "File Not Found")
        with self.assertRaises(OSError):
            with mkdir_open(path, "w") as fd:
                pass



try:
    from mock import Mock
except ImportError:
    from unittest.mock import Mock

import yaml

from passpie.importers import find_importer
from passpie.importers.default import DefaultImporter


def test_find_importer_returns_first_match_default_importer(mocker):
    mock_importer = Mock()
    mock_importer2 = Mock()
    mock_importer.match.return_value = False
    mock_importer2.match.return_value = True

    mocker.patch('passpie.importers.get_instances',
                 return_value=[mock_importer, mock_importer2])

    importer = find_importer('mockpath')

    assert importer is mock_importer2


def test_default_importer_match_passpie_exported_yaml(mocker):
    dict_content = {'handler': 'passpie', 'version': 1.0}
    mocker.patch('passpie.importers.default.DefaultImporter._read_file')
    mocker.patch('passpie.importers.default.yaml.load',
                 return_value=dict_content)

    result = DefaultImporter().match('filepath')
    assert result is True


def test_default_importer_returns_false_when_missing_version_key(mocker):
    dict_content = {'handler': 'passpie'}
    mocker.patch('passpie.importers.default.DefaultImporter._read_file')
    mocker.patch('passpie.importers.default.yaml.load',
                 return_value=dict_content)

    result = DefaultImporter().match('filepath')
    assert result is False


def test_default_importer_returns_false_when_missing_handler_key(mocker):
    dict_content = {'version': 1.0}
    mocker.patch('passpie.importers.default.DefaultImporter._read_file')
    mocker.patch('passpie.importers.default.yaml.load',
                 return_value=dict_content)

    result = DefaultImporter().match('filepath')
    assert result is False


def test_default_importer_returns_false_when_version_keys_isnt_float(mocker):
    dict_content = {'version': '1.0'}
    mocker.patch('passpie.importers.default.DefaultImporter._read_file')
    mocker.patch('passpie.importers.default.yaml.load',
                 return_value=dict_content)

    result = DefaultImporter().match('filepath')
    assert result is False


def test_default_importer_returns_loaded_credentials_from_yaml_file(mocker):
    dict_content = {'credentials': {'name': 'foo', 'name': 'bar'}}
    mocker.patch('passpie.importers.default.DefaultImporter._read_file')
    mocker.patch('passpie.importers.default.yaml.load',
                 return_value=dict_content)

    result = DefaultImporter().handle('filepath')
    assert result is dict_content.get('credentials')


def test_default_importer_match_returns_false_when_bad_yaml(mocker):
    mocker.patch('passpie.importers.default.DefaultImporter._read_file')
    mocker.patch('passpie.importers.default.yaml.load',
                 side_effect=[yaml.scanner.ScannerError])

    result = DefaultImporter().match('filepath')
    assert result is False



from .helpers import MockerTestCase
from passpie.credential import split_fullname, make_fullname


class CredentialTests(MockerTestCase):

    def test_split_fullname_raises_value_error_when_invalid_name(self):
        invalid_fullname = ""
        with self.assertRaises(ValueError):
            split_fullname(invalid_fullname)

    def test_split_fullname_returns_expected_login_and_name(self):
        self.assertEqual(split_fullname("foo@bar"), ("foo", "bar"))
        self.assertEqual(split_fullname("foo@bar.com"), ("foo", "bar.com"))
        self.assertEqual(split_fullname("@bar"), ("_", "bar"))
        self.assertEqual(
            split_fullname("foo@example.com@bar"), ("foo@example.com", "bar"))

    def test_make_fullname_returns_expected_fullname(self):
        self.assertEqual(make_fullname("foo", "bar"), "foo@bar")
        self.assertEqual(make_fullname("_", "bar"), "_@bar")
        self.assertEqual(make_fullname(None, "bar"), "_@bar")



import inspect
import os
import glob


class BaseImporter(object):
    pass


__all__ = [os.path.basename(f)[:-3]
           for f in glob.glob(os.path.dirname(__file__) + "/*.py")]


def _import_all_importer_files():
    __import__(__name__, globals(), locals(), __all__, 0)


def get_all():
    """Get all subclasses of BaseImporter from module and return and generator
    """

    _import_all_importer_files()

    for module in (value for key, value in globals().items() if key in __all__):
        for klass_name, klass in inspect.getmembers(module, inspect.isclass):
            if klass is not BaseImporter and issubclass(klass, BaseImporter):
                yield klass


def get_instances():
    """Get instances of all found detected importer classes"""
    return (wclass() for wclass in get_all())


def find_importer(filepath):
    for importer in get_instances():
        if importer.match(filepath):
            return importer



import re


def split_fullname(fullname):
    rgx = re.compile(r"(?P<login>.*)?@(?P<name>.*)")
    try:
        name = rgx.match(fullname).group("name")
        login = rgx.match(fullname).group("login")
    except AttributeError:
        raise ValueError("Not a valid name")
    return login if login else "_", name


def make_fullname(login, name):
    return "{}@{}".format("_" if login is None else login, name)



from passpie.database import PasspieStorage
from .helpers import MockerTestCase


class StorageTests(MockerTestCase):

    def setUp(self):
        self.mock_os = self.patch('passpie.database.os')
        self.mock_shutil = self.patch("passpie.database.shutil")
        self.storage = PasspieStorage("path")

    def test_read_returns_all_found_credentials_in_default_dict(self):
        mock_open = self.patch("passpie.database.open",
                               self.mock_open(), create=True)
        mock_open().read.return_value = "{}"
        self.mock_os = self.patch('passpie.database.os')
        self.mock_os.walk.return_value = [
            ('/foo', ('bar',), ('baz',)),
            ('/foo/bar', (), ('eggs.pass',)),
            ('/foo/bar2', (), ('spam.pass',))
        ]
        storage = PasspieStorage("path")
        storage.write = self.Mock()
        elements = storage.read()
        self.assertIn("_default", elements.keys())
        self.assertEqual(len(elements["_default"]), 2)

    def test_write_writes_all_elements_from_data_default_dict(self):
        mock_mkdir_open = self.patch("passpie.database.mkdir_open",
                                     self.mock_open(), create=True)

        self.patch("passpie.database.PasspieStorage.read",
                   return_value={"_default": {}})
        data = {"_default": {1: {"name": "example", "login": "foo"},
                             2: {"name": "example", "login": "bar"}}}
        storage = PasspieStorage("path")

        storage.write(data)

        self.assertEqual(mock_mkdir_open.call_count, 2)

    def test_delete_removes_credentials_files(self):
        credentials = [
            {"name": "bar", "login": "foo"},
            {"name": "spam", "login": "foozy"},
        ]

        self.storage.delete(credentials)

        for cred in credentials:
            credpath = self.mock_os.path.join(cred["name"], cred["login"])
            self.mock_os.remove.assert_any_call(credpath)

    def test_delete_removes_empty_directories(self):
        credentials = [
            {"name": "bar", "login": "foo"},
            {"name": "spam", "login": "foozy"},
        ]
        self.mock_os.listdir.side_effect = [[], ["not empty"]]
        self.storage.delete(credentials)

        credpath = self.mock_os.path.join(credentials[0]["name"],
                                          credentials[0]["login"])
        self.mock_shutil.rmtree.assert_called_once_with(
            self.mock_os.path.dirname(credpath))



from argparse import Namespace
from datetime import datetime, timedelta
from pkg_resources import get_distribution, DistributionNotFound
import functools
import json
import os
import shutil

from tabulate import tabulate
from tinydb.queries import where
import click
import pyperclip
import yaml

from ._compat import FileExistsError
from .credential import split_fullname, make_fullname
from .crypt import Cryptor
from .database import Database
from .utils import genpass


try:
    _dist = get_distribution('passpie')
    dist_loc = os.path.normcase(_dist.location)
    here = os.path.normcase(__file__)
    if not here.startswith(os.path.join(dist_loc, 'passpie')):
        raise DistributionNotFound
except DistributionNotFound:
    __version__ = 'Please install this project with setup.py or pip'
else:
    __version__ = _dist.version


CONFIG_PATH = os.path.expanduser('~/.passpierc')

config_dict = {}
if os.path.exists(CONFIG_PATH):
    with open(CONFIG_PATH) as f:
        config_content = f.read()

    try:
        config_dict = yaml.load(config_content)
        keys = ('path', 'show_password', 'headers', 'colors', 'table_format')
        for k in keys:
            assert k in config_dict
    except (AssertionError, yaml.scanner.ScannerError) as e:
        click.ClickException('Bad configuration file: {}'.format(e))

config = Namespace(
    path=config_dict.get('path', os.path.expanduser("~/.passpie")),
    show_password=config_dict.get('show_password', False),
    short_commands=config_dict.get('short_commands', False),
    table_format=config_dict.get('table_format', "fancy_grid"),
    headers=config_dict.get(
        'headers',
        ["name", "login", "password", "comment"]),
    colors=config_dict.get(
        'colors',
        {"name": "yellow", "login": "green", "password": "magenta"}),
)


class AliasedGroup(click.Group):

    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx)
                   if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail('Too many matches: %s' % ', '.join(sorted(matches)))


class Table(object):
    """Reprensent table outputs"""

    def __init__(self, data):
        self.data = data
        self.tablefmt = config.table_format

    def __getitem__(self, keys):
        if isinstance(keys, str):
            keys = (keys,)
        return [[e[k] for k in keys]
                for e in self.data]

    def render(self, keys):
        return tabulate(self[keys],
                        headers=[k.title() for k in keys],
                        tablefmt=self.tablefmt)


def get_credential_or_abort(db, fullname):
    credential = db.get(where("fullname") == fullname)
    if not credential:
        message = "Credential '{}' not found".format(fullname)
        raise click.ClickException(click.style(message, fg='red'))
    return credential


def ensure_passphrase(db, passphrase):
    try:
        with Cryptor(db._storage.path) as cryptor:
            cryptor.check(passphrase, ensure=True)
        return passphrase
    except ValueError:
        message = 'Wrong passphrase'
        raise click.ClickException(click.style(message, fg='red'))


def print_table(credentials):
    if credentials:
        for credential in credentials:
            credential['password'] = "*****"
            headers = config.headers

        for header, color in config.colors.items():
            for credential in credentials:
                credential[header] = click.style(credential[header], fg=color)

        for credential in [c for c in credentials if 'name' in c.keys()]:
            credential['name'] = click.style(credential['name'], bold=True)

        for credential in [c for c in credentials if 'login' in c.keys()]:
            credential['login'] = click.style(credential['login'], bold=True)

        click.echo(Table(credentials).render(headers))


def ensure_database(func):
    @functools.wraps(func)
    def decorator(*args, **kwargs):
        keys_path = os.path.join(config.path, '.keys')
        if os.path.isdir(config.path) and os.path.isfile(keys_path):
            return func(*args, **kwargs)
        else:
            message = 'Not initialized database at {.path}'.format(config)
            raise click.ClickException(click.style(message, fg='yellow'))
    return decorator


@click.group(cls=AliasedGroup if config.short_commands else click.Group,
             invoke_without_command=True)
@click.option('-D', '--database', help='Alternative database path',
              type=click.Path(dir_okay=True, writable=True, resolve_path=True))
@click.version_option(version=__version__)
@click.pass_context
def cli(ctx, database=config.path):
    if database:
        config.path = database
    if ctx.invoked_subcommand is None:
        db = Database(config.path)
        credentials = sorted(db.all(), key=lambda x: x["name"]+x["login"])
        print_table(credentials)


@cli.command(name='config', help='Show configuration')
def print_config():
    print(yaml.dump(vars(config), default_flow_style=False))


@cli.command(help="Initialize new passpie database")
@click.option('--passphrase', prompt=True, hide_input=True,
              confirmation_prompt=True)
@click.option('--force', is_flag=True, help="force overwrite database")
def init(passphrase, force):
    if force:
        shutil.rmtree(config.path)

    try:
        with Cryptor(config.path) as cryptor:
            cryptor.create_keys(passphrase)
    except FileExistsError:
        message = "Database exists in {}. `--force` to overwrite".format(
            config.path)
        raise click.ClickException(click.style(message, fg='yellow'))
    click.echo("Initialized database in {}".format(config.path))


@cli.command(help="Add new credential")
@click.argument("fullname")
@click.option('-r', '--random', 'password', flag_value=genpass())
@click.password_option(help="Credential password")
@click.option('-c', '--comment', default="", help="Credential comment")
@ensure_database
def add(fullname, password, comment):
    db = Database(config.path)
    try:
        login, name = split_fullname(fullname)
    except ValueError:
        message = 'invalid fullname syntax'
        raise click.ClickException(click.style(message, fg='yellow'))

    found = db.get((where("login") == login) & (where("name") == name))
    if not found:
        with Cryptor(config.path) as cryptor:
            encrypted = cryptor.encrypt(password)

        credential = dict(fullname=fullname,
                          name=name,
                          login=login,
                          password=encrypted,
                          comment=comment,
                          modified=datetime.now())
        db.insert(credential)
    else:
        message = "Credential {} already exists. --force to overwrite".format(
            fullname)
        raise click.ClickException(click.style(message, fg='yellow'))


@cli.command(help="Update credential")
@click.argument("fullname")
@click.option("--name", help="Credential new name")
@click.option("--login", help="Credential new login")
@click.option("--comment", help="Credential new comment")
@click.option("--password", help="Credential new password")
@click.option('--random', 'password', flag_value=genpass(),
              help="Credential new randomly generated password")
def update(fullname, name, login, password, comment):
    db = Database(config.path)
    credential = get_credential_or_abort(db, fullname)
    values = credential.copy()

    if any([name, login, password, comment]):
        values["name"] = name if name else credential["name"]
        values["login"] = login if login else credential["login"]
        values["password"] = password if password else credential["password"]
        values["comment"] = comment if comment else credential["comment"]
    else:
        values["name"] = click.prompt("Name", default=credential["name"])
        values["login"] = click.prompt("Login", default=credential["login"])
        values["password"] = click.prompt("Password",
                                          hide_input=True,
                                          default=credential["password"],
                                          confirmation_prompt=True,
                                          show_default=False,
                                          prompt_suffix=" [*****]: ")
        values["comment"] = click.prompt("Comment",
                                         default=credential["comment"])

    if values != credential:
        values["fullname"] = make_fullname(values["login"], values["name"])
        values["modified"] = datetime.now()
        if values["password"] != credential["password"]:
            with Cryptor(config.path) as cryptor:
                values["password"] = cryptor.encrypt(password)
        db = Database(config.path)
        db.update(values, (where("fullname") == credential["fullname"]))


@cli.command(help="Remove credential")
@click.argument("fullname")
def remove(fullname):
    db = Database(config.path)
    credential = get_credential_or_abort(db, fullname)
    if credential:
        click.confirm(
            'Remove credential: {}'.format(click.style(fullname, 'yellow')),
            abort=True
        )
        db.remove(where('fullname') == credential["fullname"])


@cli.command(help="Copy credential password to clipboard")
@click.argument("fullname")
@click.option("--passphrase", prompt="Passphrase", hide_input=True)
def copy(fullname, passphrase):
    db = Database(config.path)
    credential = get_credential_or_abort(db, fullname)
    ensure_passphrase(db, passphrase)
    with Cryptor(config.path) as cryptor:
        decrypted = cryptor.decrypt(credential["password"],
                                    passphrase=passphrase)
        pyperclip.copy(decrypted)
        click.secho("Password copied to clipboard".format(
            credential["fullname"]), fg="yellow")


@cli.command(help="Search credentials by regular expressions")
@click.argument("regex")
def search(regex):
    db = Database(config.path)
    credentials = db.search(
        where("name").matches(regex) |
        where("login").matches(regex) |
        where("comment").matches(regex))
    credentials = sorted(credentials, key=lambda x: x["name"]+x["login"])
    print_table(credentials)


@cli.command(help="Diagnose database for improvements")
@click.option("--full", is_flag=True, help="Show all entries")
@click.option("--days", default=90, type=int, help="Elapsed days")
@click.option("--passphrase", prompt="Passphrase", hide_input=True)
def status(full, days, passphrase):
    db = Database(config.path)
    ensure_passphrase(db, passphrase)
    credentials = sorted(db.all(), key=lambda x: x["name"]+x["login"])
    with Cryptor(config.path) as cryptor:
        for cred in credentials:
            cred["password"] = cryptor.decrypt(cred["password"], passphrase)

    if credentials:
        # check passwords
        for cred in credentials:
            repeated = [
                c["fullname"] for c in credentials
                if c["password"] == cred["password"] and c != cred
            ]
            if repeated:
                password = "Same as: {}".format(repeated)
                cred["repeated"] = click.style(password, "red")
            else:
                cred["repeated"] = click.style("OK", "green")

        for cred in credentials:
            cred['password'] = cred['repeated']

        # check modified time
        for cred in credentials:
            modified_delta = (datetime.now() - cred["modified"])
            if modified_delta > timedelta(days=days):
                modified_time = "{} days ago".format(modified_delta.days)
                cred["modified"] = click.style(modified_time, "red")

            else:
                cred["modified"] = click.style("OK", "green")

        if not full:
            credentials = [c for c in credentials
                           if c["password"] or c["modified"]]

        table = Table(credentials)
        click.echo(table.render(["name", "login", "password", "modified"]))


@cli.command(name="export", help="Export credentials in plain text")
@click.argument("dbfile", type=click.File("w"))
@click.option("--json", "as_json", is_flag=True, help="Export as JSON")
@click.option("--passphrase", prompt="Passphrase", hide_input=True)
def export_database(dbfile, as_json, passphrase):
    db = Database(config.path)
    ensure_passphrase(db, passphrase)
    credentials = sorted(db.all(), key=lambda x: x["name"]+x["login"])
    with Cryptor(config.path) as cryptor:
        for cred in credentials:
            cred["password"] = cryptor.decrypt(cred["password"], passphrase)

    if as_json:
        for cred in credentials:
            cred["modified"] = str(cred["modified"])
        dict_content = {
            'handler': 'passpie',
            'version': 1.0,
            'credentials': [dict(x) for x in credentials],
        }
        content = json.dumps(dict_content, indent=2)
    else:
        dict_content = {
            'handler': 'passpie',
            'version': 1.0,
            'credentials': [dict(x) for x in credentials],
        }
        content = yaml.dump(dict_content, default_flow_style=False)

    dbfile.write(content)


@cli.command(name="import", help="Import credentials from path")
@click.argument("path", type=click.Path())
def import_database(path):
    from passpie.importers import find_importer
    importer = find_importer(path)
    credentials = importer.handle(path)
    if credentials:
        db = Database(config.path)
    with Cryptor(config.path) as cryptor:
        for cred in credentials:
            encrypted = cryptor.encrypt(cred['password'])
            cred['password'] = encrypted
        db.insert_multiple(credentials)


@cli.command(help='Renew passpie password and re-encrypt credentials')
@click.option("--passphrase", prompt="Passphrase", hide_input=True)
def reset(passphrase):
    db = Database(config.path)
    ensure_passphrase(db, passphrase)
    new_passphrase = click.prompt('New passphrase',
                                  hide_input=True, confirmation_prompt=True)

    credentials = db.all()
    if credentials:
        with Cryptor(config.path) as cryptor:
            # decrypt passwords
            for cred in credentials:
                cred["password"] = cryptor.decrypt(cred["password"],
                                                   passphrase)

            # recreate keys
            cryptor.create_keys(new_passphrase, overwrite=True)

            # encrypt passwords
            for cred in credentials:
                cred["password"] = cryptor.encrypt(cred["password"])

        # remove old and insert re-encrypted credentials
        db.purge()
        db.insert_multiple(credentials)



import os

from passpie.crypt import Cryptor, KEY_INPUT
from passpie._compat import which, FileNotFoundError, FileExistsError
from .helpers import MockerTestCase


class CryptTests(MockerTestCase):

    def setUp(self):
        self.mock_gnupg = self.patch("passpie.crypt.gnupg")
        self.mock_tempfile = self.patch("passpie.crypt.tempfile")
        self.mock_shutil = self.patch("passpie.crypt.shutil")

    def test_cryptor_initialization(self):
        path = "path/to/database"
        cryptor = Cryptor(path)
        self.assertEqual(cryptor.path, path)
        self.assertEqual(cryptor.keys_path, os.path.join(path, ".keys"))
        self.assertEqual(cryptor._binary, which("gpg"))
        self.assertEqual(cryptor._homedir, self.mock_tempfile.mkdtemp())
        self.assertEqual(cryptor._gpg, self.mock_gnupg.GPG())

    def test_cryptor__enter__returns_self(self):
        cryptor = Cryptor("path/to/database")
        self.assertEqual(Cryptor.__enter__(cryptor), cryptor)

    def test_cryptor__exit__removes_homedir_temp_directory(self):
        with Cryptor("path/to/database") as cryptor:
            homedir = cryptor._homedir
        self.mock_shutil.rmtree.assert_called_once_with(homedir)

    def test_import_keys_opens_keys_path_and_import_to_gpg(self):
        to_patch = 'passpie.crypt.open'
        mock_open = self.patch(to_patch, self.mock_open(), create=True)
        mock_open().read.return_value = "key data"

        cryptor = Cryptor("path/to/database")
        cryptor._import_keys()
        cryptor._gpg.import_keys.assert_called_once_with(
            mock_open().read()
        )

    def test_import_keys_raises_file_not_found_when_keys_not_found(self):
        to_patch = 'passpie.crypt.open'
        mock_open = self.patch(to_patch, self.mock_open(), create=True)
        cryptor = Cryptor("path/to/database")

        mock_open.side_effect = OSError(2, "File Not Found")
        with self.assertRaises(FileNotFoundError):
            cryptor._import_keys()

        mock_open.side_effect = OSError(17, "File exists")
        try:
            cryptor._import_keys()
        except Exception as exc:
            self.assertNotIsInstance(exc, FileNotFoundError)

    def test_key_returns_current_key_fingerprint(self):
        cryptor = Cryptor("path/to/database")
        curkey = cryptor._gpg.list_keys().curkey

        self.assertEqual(cryptor.current_key, curkey["fingerprint"])

    def test_create_keys_raises_file_exist_error_when_keys_found_in_path(self):
        cryptor = Cryptor("path/to/database")
        self.patch("passpie.crypt.os.path.exists",
                   self.Mock(return_value=True))

        with self.assertRaises(FileExistsError):
            cryptor.create_keys("passphrase")

    def test_create_keys_export_public_and_secret_keys_into_keyspath(self):
        cryptor = Cryptor("path/to/database")
        passphrase = "passphrase"
        self.patch("passpie.crypt.os.path.exists",
                   self.Mock(return_value=False))
        keys = self.Mock(fingerprint="HEX")
        cryptor._gpg.gen_key.return_value = keys
        mock_open = self.patch("passpie.crypt.mkdir_open",
                               self.mock_open(), create=True)
        mock_keyfile = self.Mock()
        mock_open().__enter__.return_value = mock_keyfile

        cryptor.create_keys(passphrase)

        key_input = KEY_INPUT.format(passphrase)
        cryptor._gpg.gen_key.assert_called_once_with(key_input)
        cryptor._gpg.export_keys.assert_any_call(keys.fingerprint)

        pubkey = cryptor._gpg.export_keys(keys.fingerprint)
        seckey = cryptor._gpg.export_keys(keys.fingerprint, secret=True)
        cryptor._gpg.export_keys.assert_any_call(keys.fingerprint, secret=True)
        mock_keyfile.write.called_once_with(pubkey + seckey)

    def test_encrypt_returns_gpg_encrypted_data(self):
        cryptor = Cryptor("path/to/database")
        cryptor._import_keys = self.Mock()
        encrypted = "encrypted data"
        cryptor._gpg.encrypt.return_value = encrypted

        self.assertEqual(cryptor.encrypt("data"), str(encrypted))

    def test_decrypt_returns_gpg_decrypted_data(self):
        cryptor = Cryptor("path/to/database")
        cryptor._import_keys = self.Mock()
        decrypted = "decrypted data"
        cryptor._gpg.decrypt.return_value = decrypted

        self.assertEqual(cryptor.decrypt("data", "passphrase"), str(decrypted))

    def test_passphrase_check_raises_value_error_when_bad_passphrase(self):
        cryptor = Cryptor("path/to/database")
        cryptor._import_keys = self.Mock()
        passphrase = "passphrase"
        result = cryptor.check(passphrase)

        self.assertTrue(result)
        self.assertTrue(cryptor._gpg.sign.called)
        cryptor._gpg.sign.assert_called_once_with(
            "testing",
            default_key=cryptor.current_key,
            passphrase=passphrase
        )

    def test_passprase_check_with_ensure_raises_value_error_with_ensure(self):
        cryptor = Cryptor("path/to/database")
        cryptor._import_keys = self.Mock()
        cryptor._gpg.sign.return_value = None
        passphrase = "passphrase"

        with self.assertRaises(ValueError):
            cryptor.check(passphrase, ensure=True)



from functools import partial
import os
import shutil

from tinydb import TinyDB, Storage
import yaml
from .utils import mkdir_open


class PasspieStorage(Storage):
    extension = ".pass"

    def __init__(self, path):
        super(PasspieStorage, self).__init__()
        self.path = path

    def delete(self, credentials):
        for cred in credentials:
            dirname, filename = cred["name"], cred["login"] + self.extension
            credpath = os.path.join(self.path, dirname, filename)
            os.remove(credpath)
            if not os.listdir(os.path.dirname(credpath)):
                shutil.rmtree(os.path.dirname(credpath))

    def read(self):
        elements = []
        for rootdir, dirs, files in os.walk(self.path):
            filenames = [f for f in files if f.endswith(self.extension)]
            for filename in filenames:
                docpath = os.path.join(rootdir, filename)
                with open(docpath) as f:
                    elements.append(yaml.load(f.read()))

        return {"_default":
                {idx: elem for idx, elem in enumerate(elements, start=1)}}

    def write(self, data):
        deleted = [c for c in self.read()["_default"].values()
                   if c not in data["_default"].values()]
        self.delete(deleted)

        for eid, cred in data["_default"].items():
            dirname, filename = cred["name"], cred["login"] + self.extension
            credpath = os.path.join(self.path, dirname, filename)
            with mkdir_open(credpath, "w") as f:
                f.write(yaml.dump(dict(cred), default_flow_style=False))


Database = partial(TinyDB, storage=PasspieStorage)



from click.testing import CliRunner
from tinydb import TinyDB
from tinydb.storages import MemoryStorage
import pytest
try:
    import mock
except ImportError:
    from unittest import mock

from passpie import cli


@pytest.fixture
def mock_db(mocker):
    credentials = [
        {'login': 'foo', 'name': 'bar', 'fullname': 'foo@bar',
         'password': '', 'comment': ''},
        {'login': 'foa', 'name': 'bazzy', 'fullname': 'foa@bazzy',
         'password': '', 'comment': ''},
        {'login': 'spam', 'name': 'egg', 'fullname': 'spam@egg',
         'password': '', 'comment': ''},
    ]
    database = TinyDB(storage=MemoryStorage)
    database.insert_multiple(credentials)
    MockDB = mock.MagicMock(return_value=database)
    mocker.patch('passpie.cli.Database', MockDB)
    return database


def test_cli_search_find_results_by_login_regex(mock_db):
    runner = CliRunner()
    result = runner.invoke(cli.search, ['fo[oa]'])

    assert result.exit_code == 0
    assert 'foo' in result.output
    assert 'foa' in result.output
    assert 'spam' not in result.output


def test_cli_remove_delete_credential_found_by_database(mock_db):
    runner = CliRunner()
    result = runner.invoke(cli.remove, ['foo@bar'], input='y')
    result_print = runner.invoke(cli.cli)

    assert result.exit_code == 0
    assert 'foo' not in result_print.output


def test_cli_copy_credential_password_to_database(mocker, mock_db):
    fullname = 'foo@bar'
    password = 's3cr3t'
    mocker.patch('passpie.cli.ensure_passphrase')
    mock_pyperclip = mocker.patch('passpie.cli.pyperclip')
    mocker.patch('passpie.cli.Cryptor.decrypt',
                 mock.Mock(return_value=password))
    runner = CliRunner()
    result = runner.invoke(cli.copy, [fullname], input='passphrase')

    assert result.exit_code == 0
    assert mock_pyperclip.copy.called
    mock_pyperclip.copy.assert_called_once_with(password)


def test_cli_reset_database_overwrite_old_keys(mocker, mock_db):
    mocker.patch('passpie.cli.ensure_passphrase')
    mock_create_keys = mocker.patch('passpie.cli.Cryptor.create_keys')
    mock_decrypt = mocker.patch('passpie.cli.Cryptor.decrypt')
    mock_decrypt.return_value = ''
    new_passphrase = mocker.patch('passpie.cli.click.prompt')()

    runner = CliRunner()
    result = runner.invoke(cli.reset, input='passphrase')

    assert result.exit_code == 0
    assert mock_create_keys.called
    mock_create_keys.assert_called_once_with(new_passphrase, overwrite=True)


def test_cli_reset_database_re_encrypt_all_passwords(mocker, mock_db):
    password = 's3cr3t'
    mocker.patch('passpie.cli.ensure_passphrase')
    mocker.patch('passpie.cli.Cryptor.create_keys')
    mocker.patch('passpie.cli.click.prompt')
    mock_encrypt = mocker.patch('passpie.cli.Cryptor.encrypt')
    mock_decrypt = mocker.patch('passpie.cli.Cryptor.decrypt')
    mock_decrypt.return_value = password

    runner = CliRunner()
    result = runner.invoke(cli.reset, input='passphrase')

    assert result.exit_code is 0
    assert mock_encrypt.call_count is len(mock_db.all())
    assert mock_decrypt.call_count is len(mock_db.all())
    for credential in mock_db.all():
        mock_decrypt.assert_any_call_with(credential['password'])
    for credential in mock_db.all():
        mock_encrypt.assert_any_call_with(password)


def test_cli_reset_purges_all_elements(mocker, mock_db):
    mocker.patch('passpie.cli.ensure_passphrase')
    mocker.patch('passpie.cli.Cryptor.create_keys')
    mocker.patch('passpie.cli.click.prompt')
    mocker.patch('passpie.cli.Cryptor.encrypt')
    mocker.patch('passpie.cli.Cryptor.decrypt')
    mocker.patch.object(mock_db, 'purge')

    runner = CliRunner()
    result = runner.invoke(cli.reset, input='passphrase')

    assert result.exit_code is 0
    assert mock_db.purge.called



import yaml
from yaml.scanner import ScannerError

from passpie.importers import BaseImporter


class DefaultImporter(BaseImporter):

    def _read_file(filepath):
        return open(filepath).read()

    def match(self, filepath):
        try:
            dict_content = yaml.load(self._read_file(filepath))
        except ScannerError:
            return False

        try:
            assert dict_content.get('handler') == 'passpie'
            assert isinstance(dict_content.get('version'), float)
        except AssertionError:
            return False

        return True

    def handle(self, filepath):
        dict_content = yaml.load(self._read_file(filepath))
        credentials = dict_content.get('credentials')
        return credentials



import sys
try:
    from shutil import which
except ImportError:
    from distutils.spawn import find_executable as which


def is_python2():
    return sys.version_info < (3,)


class FileNotFoundError(OSError):

    def __init__(self, message="No such file or directory"):
        super(FileNotFoundError, self).__init__(2, message)


class FileExistsError(OSError):

    def __init__(self, message="File exists"):
        super(FileExistsError, self).__init__(17, message)



from contextlib import contextmanager
import errno
import os
import random
import string


def genpass(length=32, special="_-#|+="):
    """generates a password with random chararcters
    """
    chars = special + string.ascii_letters + string.digits + " "
    return "".join(random.choice(chars) for _ in range(length))


@contextmanager
def mkdir_open(path, mode="r"):
    try:
        dir_path = os.path.dirname(path)
        os.makedirs(dir_path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(dir_path):
            pass
        else:
            raise
    with open(path, mode) as fd:
        yield fd



#!/usr/bin/env python
from setuptools import setup, find_packages


__version__ = "0.1rc3"

requirements_file = "requirements.txt"
requirements = [pkg.strip() for pkg in open(requirements_file).readlines()]
requirements_test_file = "requirements_test.txt"
requirements_tests = [pkg.strip() for pkg in open(requirements_file).readlines()]

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst') + "\n"
except(IOError, ImportError):
    long_description = open('README.md').read() + "\n"

setup(
    name='passpie',
    version=__version__,
    license='License :: OSI Approved :: MIT License',
    description="Manage your login credentials from the terminal painlessly.",
    long_description=long_description,
    author='Marcwebbie',
    author_email='marcwebbie@gmail.com',
    url='https://github.com/marcwebbie/passpie',
    download_url='https://pypi.python.org/pypi/passpie',
    packages=find_packages(),
    entry_points="""
        [console_scripts]
        passpie=passpie.cli:cli
    """,
    install_requires=requirements,
    tests_require=requirements_tests,
    test_suite='tests',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Programming Language :: Python',
        'Topic :: Security :: Cryptography',
    ],
)



import errno
import os
import shutil
import tempfile
import gnupg

from ._compat import which, FileNotFoundError, FileExistsError
from .utils import mkdir_open


KEY_INPUT = """Key-Type: RSA
Key-Length: 4096
Subkey-Type: RSA
Name-Comment: Auto-generated by Passpie
Passphrase: {}
Name-Real: Passpie
Name-Email: passpie@local
Expire-Date: 0
%commit
"""


class Cryptor(object):

    def __init__(self, path):
        self.path = path
        self.keys_path = os.path.join(path, ".keys")
        self._binary = which("gpg")
        self._homedir = tempfile.mkdtemp()
        self._gpg = gnupg.GPG(binary=self._binary, homedir=self._homedir)

    def __enter__(self):
        return self

    def __exit__(self, exc_ty, exc_val, exc_tb):
        shutil.rmtree(self._homedir)

    def _import_keys(self):
        try:
            with open(self.keys_path) as keyfile:
                self._gpg.import_keys(keyfile.read())
        except OSError as exc:
            if exc.errno == errno.ENOENT:
                raise FileNotFoundError("Keys not found in path")
            else:
                raise

    @property
    def current_key(self):
        keys = self._gpg.list_keys()
        return keys.curkey["fingerprint"]

    def create_keys(self, passphrase, overwrite=False):
        if overwrite is False and os.path.exists(self.keys_path):
            raise FileExistsError("Keys found in path")

        keys = self._gpg.gen_key(KEY_INPUT.format(passphrase))
        pubkey = self._gpg.export_keys(keys.fingerprint)
        seckey = self._gpg.export_keys(keys.fingerprint, secret=True)
        with mkdir_open(self.keys_path, "w") as keyfile:
            keyfile.write(pubkey + seckey)

    def encrypt(self, data):
        self._import_keys()
        encrypted = self._gpg.encrypt(data, self.current_key)
        return str(encrypted)

    def decrypt(self, data, passphrase):
        self._import_keys()
        self.check(passphrase, ensure=True)
        decrypted = self._gpg.decrypt(data, passphrase=passphrase)
        return str(decrypted)

    def check(self, passphrase, ensure=False):
        self._import_keys()
        sign = self._gpg.sign(
            "testing",
            default_key=self.current_key,
            passphrase=passphrase
        )
        if sign:
            return True
        else:
            if ensure:
                raise ValueError("Wrong passphrase")



import unittest

try:
    from unittest.mock import patch, Mock, MagicMock, mock_open
except ImportError:
    from mock import patch, Mock, MagicMock, mock_open


class MockerTestCase(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        self.mock_open = mock_open
        self.Mock = Mock
        self.MagicMock = MagicMock

        super(MockerTestCase, self).__init__(*args, **kwargs)

    def patch(self, *args, **kwargs):
        patcher = patch(*args, **kwargs)
        thing = patcher.start()
        self.addCleanup(patcher.stop)
        return thing

    def patch_object(self, *args, **kwargs):
        patcher = patch.object(*args, **kwargs)
        thing = patcher.start()
        self.addCleanup(patcher.stop)
        return thing

    def patch_multiple(self, *args, **kwargs):
        patcher = patch.multiple(*args, **kwargs)
        thing = patcher.start()
        self.addCleanup(patcher.stop)
        return thing

    def patch_dict(self, *args, **kwargs):
        patcher = patch.dict(*args, **kwargs)
        thing = patcher.start()
        self.addCleanup(patcher.stop)
        return thing


ENCRYPTED_DATA = """
-----BEGIN PGP MESSAGE-----

hQEMA1xKO0QntlBSAQgAnojr1QGGxd3ihH7ET0mlNkfpvb4tKLySNeKFoj+DqjI2
OxtJgD+adpalSAMQD5L57ttxPWSXzhXgnbCZslxFQWYz30j3BnaNe4cV4JMuL3qF
8/1E0xpZIBIza7I8Loo7IV7fVzKdv00T6gKImQQRgPVdTrLcmHk3xX8xEwTmdIdy
ZCjk3Hnn/i5AkXjIjxJ3NiGsAyAPZWtX91PRN2X4+GuIaN7sWMCGzzOY2roZxmNb
mMRpAb7NzxORORHbWXjH/0Y09DkaqNoZVBDN0fp2bYmFClU98ULW4NlyK6xcuTkp
2c5WMeX70maRzciXC81HnfASvFYPlgm0aquTi+/rstJSAaDQIXPCL5GQX4TQ9Vyy
ni3FsO40+4UBXUm03cLiy8wWTL2OIxs3QmVY2HsUGacypZf+r3NNwV3/4bR757qj
JmNIw2541RSIoUbitA8aw7el/w==
=VRFD
-----END PGP MESSAGE-----
"""
